/***************************************************************************
                          eqband.cpp  -  description
                             -------------------
    begin                : Mon Jun 16 20:12:07 CEST 2006
    copyright            : (C) 2006 by Nicola Doebelin
    email                : nicola@doebelin.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "eqband.h"
#include <math.h>

static const double pi=3.1415926535897932384626433832795;

EqBand::EqBand(int i)
{
  _id = i;
  _freq = 1000.0;
  _gain = 0.0;
  _mode = 0;
  _bandwidth = 1.0;
  _active = false;
  _samplerate = 44100.0;

  // reset filter coeffs
  b0a0 = b1a0 = b2a0 = a1a0 = a2a0 = 0.0;

  // reset in/out history
  ou1 = ou2 = in1 = in2 = 0.0f;
}

EqBand::~EqBand()
{}

QString EqBand::modeString()
{
  QString s;
  if (_mode == 6) s = "Peak";
  if (_mode == 0) s = "Low pass";
  if (_mode == 1) s = "High pass";
  if (_mode == 7) s = "Shelf";
  if (_mode == 8) s = "Shelf";
  return s;
}

void EqBand::calcFilterCoeffs()
{
	bool q = true;
	if (_mode <= 1) q = false;  // lowpass and highpass
	calcFilterCoeffs(_mode, (double)_freq, (double)_samplerate, (double)_bandwidth, (double)_gain, q);
}

void EqBand::calcFilterCoeffs(int const type, double const frequency, double const sample_rate, double const q, double const db_gain, bool q_is_bandwidth)
{
  // temp coef vars
  double alpha = 0.0,
            a0 = 0.0,
            a1 = 0.0,
            a2 = 0.0,
            b0 = 0.0,
            b1 = 0.0,
            b2 = 0.0;

  // peaking, lowshelf and hishelf
  if(type>=6)
  {
    double const A     = pow(10.0,(db_gain/40.0));
    double const omega = 2.0*pi*frequency/sample_rate;
    double const tsin  = sin(omega);
    double const tcos  = cos(omega);

    if(q_is_bandwidth)
      alpha=tsin*sinh(log(2.0)/2.0*q*omega/tsin);
    else
      alpha=tsin/(2.0*q);

    double const beta	=	sqrt(A)/q;

    // peaking
    if(type==6)
    {
      b0=float(1.0+alpha*A);
      b1=float(-2.0*tcos);
      b2=float(1.0-alpha*A);
      a0=float(1.0+alpha/A);
      a1=float(-2.0*tcos);
      a2=float(1.0-alpha/A);
    }

    // lowshelf
    if(type==7)
    {
      b0=float(A*((A+1.0)-(A-1.0)*tcos+beta*tsin));
      b1=float(2.0*A*((A-1.0)-(A+1.0)*tcos));
      b2=float(A*((A+1.0)-(A-1.0)*tcos-beta*tsin));
      a0=float((A+1.0)+(A-1.0)*tcos+beta*tsin);
      a1=float(-2.0*((A-1.0)+(A+1.0)*tcos));
      a2=float((A+1.0)+(A-1.0)*tcos-beta*tsin);
    }

    // hishelf
    if(type==8)
    {
      b0=float(A*((A+1.0)+(A-1.0)*tcos+beta*tsin));
      b1=float(-2.0*A*((A-1.0)+(A+1.0)*tcos));
      b2=float(A*((A+1.0)+(A-1.0)*tcos-beta*tsin));
      a0=float((A+1.0)-(A-1.0)*tcos+beta*tsin);
      a1=float(2.0*((A-1.0)-(A+1.0)*tcos));
      a2=float((A+1.0)-(A-1.0)*tcos-beta*tsin);
    }
  }
  else
  {
    // other filters
    double const omega	=	2.0*pi*frequency/sample_rate;
    double const tsin	=	sin(omega);
    double const tcos	=	cos(omega);

    if(q_is_bandwidth)
      alpha=tsin*sinh(log(2.0)/2.0*q*omega/tsin);
    else
      alpha=tsin/(2.0*q);


    // lowpass
    if(type==0)
    {
      b0=(1.0-tcos)/2.0;
      b1=1.0-tcos;
      b2=(1.0-tcos)/2.0;
      a0=1.0+alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }

    // hipass
    if(type==1)
    {
      b0=(1.0+tcos)/2.0;
      b1=-(1.0+tcos);
      b2=(1.0+tcos)/2.0;
      a0=1.0+ alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }

    // bandpass csg
    if(type==2)
    {
      b0=tsin/2.0;
      b1=0.0;
      b2=-tsin/2;
      a0=1.0+alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }

    // bandpass czpg
    if(type==3)
    {
      b0=alpha;
      b1=0.0;
      b2=-alpha;
      a0=1.0+alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }

    // notch
    if(type==4)
    {
      b0=1.0;
      b1=-2.0*tcos;
      b2=1.0;
      a0=1.0+alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }

    // allpass
    if(type==5)
    {
      b0=1.0-alpha;
      b1=-2.0*tcos;
      b2=1.0+alpha;
      a0=1.0+alpha;
      a1=-2.0*tcos;
      a2=1.0-alpha;
    }
  }

  // set filter coeffs
  b0a0=float(b0/a0);
  b1a0=float(b1/a0);
  b2a0=float(b2/a0);
  a1a0=float(a1/a0);
  a2a0=float(a2/a0);
};

float EqBand::filter(float in0)
{
  // filter
  float const yn = b0a0*in0 + b1a0*in1 + b2a0*in2 - a1a0*ou1 - a2a0*ou2;

  // push in/out buffers
  in2=in1;
  in1=in0;
  ou2=ou1;
  ou1=yn;

  // return output
  return yn;
};

void EqBand::calcCurve(QList<QPointF> &list)
{
  calcFilterCoeffs();  // update the coefficients before curve calculation

  for (int i = 0; i < list.size(); i++) {
    double w = 2.0 * pi * list[i].x() / _samplerate;
    double phi = 4.0 * pow(sin(w/2.0), 2.0);

    double val = 10.0 * log10(pow((b0a0 + b1a0 + b2a0), 2.0) + (b0a0 * b2a0 * phi -
                              (b1a0 * (b0a0 + b2a0) + 4.0 * b0a0 * b2a0)) * phi) -
                 10.0 * log10(pow(1.0 + a1a0 + a2a0, 2.0) + (1.0 * a2a0 * phi -
                              (a1a0 * (1.0 + a2a0) + 4.0 * 1.0 * a2a0)) * phi);

    list[i].setY(float(val));
  }
}


/* EOF */
