/***************************************************************************
                          jmbeq.cpp  -  description
                             -------------------
    begin                : Mon Jun 16 20:12:07 CEST 2006
    copyright            : (C) 2006 by Nicola Doebelin
    email                : nicola@doebelin.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "jmbeq.h"

//#include <QApplication>
#include <QColor>
#include <QCursor>
#include <QFontMetrics>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QPen>
#include <QString>
#include <QPainter>
#include <QPalette>
#include <QtCore>

#include <math.h>

static const bool USE_ANTIALIASING	= true;
static const bool HIDE_CURSOR_ON_ACTION	= true;
static const int LOWER_LEVEL_LIMIT	= -15;
static const int UPPER_LEVEL_LIMIT	= 15;
static const int LOWER_FREQ_LIMIT	= 20;
static const int UPPER_FREQ_LIMIT	= 20000;
static const float UPPER_Q_LIMIT	= 6.0;
static const float LOWER_Q_LIMIT	= 0.3;
static const int VMARGIN		= 20;
static const int RMARGIN		= 15;
static const float CURSOR_SPEED		= 50.0;
static const QString VERSION		= "0.0.5";
static const int BUTTON_DIAMETER	= 10;
static const QString ACTIVE_COLOR	= "#52507b";
static const QString DOT_COLOR		= "#cddeff";
static const QString BG_COLOR		= "#ffffff";
static const QString INACTIVE_COLOR	= "#cddeff";
static const QString FREQ_KEY		= "f";
static const QString GAIN_KEY		= "g";
static const QString DRAG_KEY		= "d";
static const QString BW_KEY		= "q";
static const QString MODE_KEY		= "m";
static const QString RESET_KEY		= "r";
static const QString FILTER_KEYS	= "1;2;3;4;5;6;7;8;9";

JmbEq::JmbEq(QWidget *parent, Qt::WFlags f)
    : QDialog(parent, f)
{
	qDebug() << "JmbEq version" << VERSION;

	setWindowTitle("JmbEq " + VERSION);
	filterKeys = FILTER_KEYS.split(";");

	setAutoFillBackground(true);
	QPalette pal = palette();
	pal.setColor(QPalette::Window, QColor(BG_COLOR));
	setPalette(pal);

	setMouseTracking(true);
	hasActiveBand = false;
	activeAction = NONE;
	ignoreMove = false;
	oGain = 0.0;

	helpKeys << QString("Band: <1..4>") << QString("Drag: [" + DRAG_KEY + "]")
		<< QString("Gain: [" + GAIN_KEY + "]") << QString("Freq: [" + FREQ_KEY + "]")
		<< QString("Q: [" + BW_KEY + "]") << QString("Reset: <" + RESET_KEY + ">");

	QList<float> defaultFreqs;
	defaultFreqs << 40.0 << 500.0 << 2000.0 << 12000.0;

	// create 4 filter bands
	for (int i = 0; i < 4; i++) {
		EqBand filter(i);
		filter.setMode(6);
		filter.setFreq(defaultFreqs[i]);
		filterList.push_back(filter);
	}

	// frequencies used to draw the eq curve
	for (int i = 0; i < 3; i++) {
		for (int j = 0; j < 179; j++)
			curvePoints.append(QPointF(pow(10.0, float(i)) * (20.0 + j * 1.0), 0.0));
	}
	curvePoints.append(QPointF(20000.0, 0.0));

	calcEqCurve();
}

JmbEq::~JmbEq()
{
	qDebug() << "Exiting JmbEq";
}

void JmbEq::resizeEvent(QResizeEvent *)
{
	calcEqCurve();
	update();
}

// drawing directly on the widget, no double buffering required in Qt4
void JmbEq::paintEvent(QPaintEvent *)
{
	drawGrid();
	drawText();

	QPainter p(this);
		QFontMetrics ftm(p.font());
		int dist = ftm.descent() + 3;

		if (USE_ANTIALIASING) {
			p.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing, true);
		} else {
			p.setRenderHints(QPainter::Antialiasing | QPainter::TextAntialiasing, false);
		}

		// draw the eq dots and text
		for (int i = 0; i < filterList.size(); i++) {
			QString s(QString("%1").arg(i+1));		// label
			int x = freqToX(filterList[i].freq());
			int y = levelToY(filterList[i].gain());
			int xx = x - ftm.width(s)/2;			// simplifies drawing of background
			int yy = y - BUTTON_DIAMETER/2 - dist;		// rectangles for the channel numbers

			// draw background rectangle
			p.setBrush(QColor(BG_COLOR));
			p.setPen(QColor(BG_COLOR));
			p.drawRect(xx - 1, yy - ftm.ascent() - 1, ftm.width(s) + 1, ftm.ascent() + 1);

			if (filterList[i].isActive()) {
				p.setBrush(QColor(ACTIVE_COLOR));
				p.setPen(QColor(ACTIVE_COLOR));
			} else {
				p.setBrush(QColor(INACTIVE_COLOR));
				p.setPen(QColor(INACTIVE_COLOR));
			}

			// draw the label
			p.drawText(xx, yy, s);

			// draw the dot
			p.drawEllipse(x - BUTTON_DIAMETER/2, y - BUTTON_DIAMETER/2,
			              BUTTON_DIAMETER, BUTTON_DIAMETER);
		}

		// draw the eq curve
		QColor color(ACTIVE_COLOR);
		QPen pen(color);
		pen.setWidth(2);
		p.setPen(pen);

		if (USE_ANTIALIASING) {
			for (int i = 1; i < (int)curvePoints.size(); i++) {
				p.drawLine(	QPointF(f_freqToX(curvePoints[i-1].x()),
							f_levelToY(curvePoints[i-1].y())), 
		        	   		QPointF(f_freqToX(curvePoints[i].x()),  
							f_levelToY(curvePoints[i].y())));
			}
		} else {
			for (int i = 1; i < (int)curvePoints.size(); i++) {
				p.drawLine(	int(0.5 + freqToX(curvePoints[i-1].x())),
						int(0.5 + levelToY(curvePoints[i-1].y())), 
		        	   		int(0.5 + freqToX(curvePoints[i].x())),  
						int(0.5 + levelToY(curvePoints[i].y())));
			}
		}

		// draw the output gain level indicator
		int b = levelToY(0.0);
		int t = levelToY(outputGain());
		p.setPen(Qt::NoPen);
		if (activeAction == MASTERGAIN) p.setBrush(QColor(ACTIVE_COLOR));
		else                            p.setBrush(QColor(INACTIVE_COLOR));
		p.drawRect(width()-RMARGIN+3, b, RMARGIN-5, t-b);
}

void JmbEq::mouseMoveEvent(QMouseEvent *e)
{
	if ((activeAction == NONE) || ignoreMove) {
		e->ignore();
		return;
	}

	// keep the cursor inside the widget, otherwise the widget may lose focus
	int dx = startPos.x() - e->pos().x();
	int dy = startPos.y() - e->pos().y();

	ignoreMove = true;
	if (e->pos().x() >= width() - 1) {
		QCursor::setPos(mapToGlobal(QPoint(5, e->pos().y())));
		startPos = mapFromGlobal(QPoint(QCursor::pos().x() + dx, QCursor::pos().y() + dy));
	}
	if (e->pos().x() <= 1) {
		QCursor::setPos(mapToGlobal(QPoint(width() - 5, e->pos().y())));
		startPos = mapFromGlobal(QPoint(QCursor::pos().x() + dx, QCursor::pos().y() + dy));
	}
	if (e->pos().y() >= height() - 1) {
		QCursor::setPos(mapToGlobal(QPoint(e->pos().x(), 5)));
		startPos = mapFromGlobal(QPoint(QCursor::pos().x() + dx, QCursor::pos().y() + dy));
	}
	if (e->pos().y() <= 1) {
		QCursor::setPos(mapToGlobal(QPoint(e->pos().x(), height() - 5)));
		startPos = mapFromGlobal(QPoint(QCursor::pos().x() + dx, QCursor::pos().y() + dy));
	}
	ignoreMove = false;
	// end of keeping the cursor inside the widget

	float fx = xToFreq(prevPos.x() + e->x() - startPos.x());
	float fy = yToLevel(prevPos.y() + e->y() - startPos.y());
	if (fx > float(UPPER_FREQ_LIMIT)) fx = (float)UPPER_FREQ_LIMIT;
	if (fx < float(LOWER_FREQ_LIMIT)) fx = (float)LOWER_FREQ_LIMIT;
	if (fy > float(UPPER_LEVEL_LIMIT)) fy = (float)UPPER_LEVEL_LIMIT;
	if (fy < float(LOWER_LEVEL_LIMIT)) fy = (float)LOWER_LEVEL_LIMIT;

	if (activeAction == FREQ) {
		activeBand->setFreq(fx);
	}
	if (activeAction == GAIN) {
		activeBand->setGain(fy);
	}
	if (activeAction == MASTERGAIN) {
		setOutputGain(fy);
	}
	if (activeAction == BW) {
		float q = lastQ + float(startPos.y() - e->y()) / CURSOR_SPEED;
		if (q < LOWER_Q_LIMIT) q = LOWER_Q_LIMIT;
		if (q > UPPER_Q_LIMIT) q = UPPER_Q_LIMIT;
		activeBand->setBandwidth(q);
	}
	if (activeAction == DRAG) {
		activeBand->setFreq(fx);
		activeBand->setGain(fy);
	}

	calcEqCurve();
	update();
}

void JmbEq::keyPressEvent(QKeyEvent *e)
{
	// we are not interested in autorepeated key events
	if (e->isAutoRepeat()) {
		e->ignore();
		return;
	}

	// close the dialog
	if (e->key() == Qt::Key_Return) close();
	if (e->key() == Qt::Key_Escape) close();

	// determine current values and store mouse position when the key is pressed
	startPos = mapFromGlobal(QCursor::pos());
	cursorPos = startPos;
	if (hasActiveBand) prevPos = QPoint(freqToX(activeBand->freq()), levelToY(activeBand->gain()));

	// activate actions depending on the key pressed
	if (e->text() == FREQ_KEY) {
		if (hasActiveBand) activeAction = FREQ;
		if (HIDE_CURSOR_ON_ACTION) setCursor(QCursor(Qt::BlankCursor));
		return;
	}
	if (e->text() == GAIN_KEY) {
		if (HIDE_CURSOR_ON_ACTION) setCursor(QCursor(Qt::BlankCursor));
		if (hasActiveBand) activeAction = GAIN;
		else {
			activeAction = MASTERGAIN;
			prevPos = QPoint(0, levelToY(outputGain()));
		}
		return;
	}
	if (e->text() == BW_KEY) {
		if (HIDE_CURSOR_ON_ACTION) setCursor(QCursor(Qt::BlankCursor));
		if (hasActiveBand) {
			activeAction = BW;
			lastQ = activeBand->bandwidth();
		}
		return;
	}
	if (e->text() == DRAG_KEY) {
		if (hasActiveBand) activeAction = DRAG;
		if (HIDE_CURSOR_ON_ACTION) setCursor(QCursor(Qt::BlankCursor));
		return;
	}
	if (e->text() == RESET_KEY) {
		reset();
		return;
	}
	if (e->text() == MODE_KEY) {
		toggleMode();
		return;
	}

	// (de)activate filter band
	if (filterKeys.contains(e->text())) {
		int a = 0; // used to determine whether there is an active band

		for (int i = 0; i < (int)filterList.size(); i++) {
			if ( e->text().toInt() == (i + 1) ) {
				filterList[i].toggleActive();
				if (filterList[i].isActive()) a++;  // could also be set to inactive
					activeBand = &filterList[i];
			} else filterList[i].setActive(false);
		}

		if (a) hasActiveBand = true;
		else   hasActiveBand = false;
		update();
	}
}

void JmbEq::keyReleaseEvent(QKeyEvent *e)
{
	if (e->isAutoRepeat()) {
		e->ignore();
		return;
	}

	// whatever key it was, deactivate all key actions
	activeAction = NONE;
	QCursor::setPos(mapToGlobal(cursorPos));
	if (HIDE_CURSOR_ON_ACTION) unsetCursor();
	update();
}

// draw the background image
void JmbEq::drawGrid()
{
	QPainter p(this);
		p.setRenderHint(QPainter::Antialiasing, false);

		// horizontal lines
		p.setPen(QColor(INACTIVE_COLOR));
		for (int i = 3; i <= UPPER_LEVEL_LIMIT; i += 3)
			p.drawLine(0, levelToY(float(i)), width()-RMARGIN, levelToY(float(i))); // 3, 6, 9 dB ...

		for (int i = -3; i >= LOWER_LEVEL_LIMIT; i -= 3)
			p.drawLine(0, levelToY(float(i)), width()-RMARGIN, levelToY(float(i))); // -3, -6, -9 dB ...

		// vertical lines
		for (float i = 1.0; i <= 100.0; i *= 10.0) {
			p.drawLine(freqToX(i*20.0), VMARGIN, freqToX(i*20.0), height() - VMARGIN);
			p.drawLine(freqToX(i*40.0), VMARGIN, freqToX(i*40.0), height() - VMARGIN);
			p.drawLine(freqToX(i*60.0), VMARGIN, freqToX(i*60.0), height() - VMARGIN);
			p.drawLine(freqToX(i*80.0), VMARGIN, freqToX(i*80.0), height() - VMARGIN);
			p.drawLine(freqToX(i*100.0), VMARGIN, freqToX(i*100.0), height() - VMARGIN);
			p.drawLine(freqToX(i*120.0), VMARGIN, freqToX(i*120.0), height() - VMARGIN);
			p.drawLine(freqToX(i*140.0), VMARGIN, freqToX(i*140.0), height() - VMARGIN);
			p.drawLine(freqToX(i*160.0), VMARGIN, freqToX(i*160.0), height() - VMARGIN);
			p.drawLine(freqToX(i*180.0), VMARGIN, freqToX(i*180.0), height() - VMARGIN);
		}

		p.drawLine(freqToX(20000.0), VMARGIN, freqToX(20000.0), height() - VMARGIN);
		p.drawLine(0, levelToY(0.0), width(), levelToY(0.0)); // 0 dB line
}

void JmbEq::drawText()
{
	QPainter p(this);
		QFontMetrics ftm(p.font());
		int distb = ftm.descent() + 3;
		int distt = ftm.ascent() + 3;
		p.setRenderHint(QPainter::Antialiasing, false);

		int tlen = ftm.width(helpKeys.join(""));
		// calculate spaces for regular distribution
		int spc = int(float(width() - tlen - 10)/float(helpKeys.size() - 1) + 0.5);
		int left = 5;

		p.setPen(QColor(INACTIVE_COLOR));
		for (int i = 0; i < helpKeys.size(); i++) {
			p.drawText(left, distt, helpKeys[i]);
			left += ftm.width(helpKeys[i]) + spc;
		}

		p.setPen(QColor(ACTIVE_COLOR));
		p.drawText(4*width()/5, height() - distb, QString("Output: %1").arg(outputGain(), 0, 'f', 2));

		if (hasActiveBand) {
			p.drawText(          5, height() - distb, QString("Gain: %1 dB").arg(activeBand->gain(), 0, 'f', 2));
			p.drawText(1*width()/5, height() - distb, QString("Freq: %1 Hz").arg(activeBand->freq(), 0, 'f', 0));
			p.drawText(2*width()/5, height() - distb, QString("Q: %1").arg(activeBand->bandwidth(), 0, 'f', 2));
			p.drawText(3*width()/5, height() - distb, QString("Mode: %1").arg(activeBand->modeString()));
		} else {
			p.setPen(QColor(INACTIVE_COLOR));
			p.drawText(          5, height() - distb, QString("Gain:"));
			p.drawText(1*width()/5, height() - distb, QString("Freq:"));
			p.drawText(2*width()/5, height() - distb, QString("Q:"));
			p.drawText(3*width()/5, height() - distb, QString("Mode:"));
		}

}

// calculate the eq curve to be displayed
void JmbEq::calcEqCurve()
{
	// create a copy of the curve point list
	QList<QPointF> tempCurve = curvePoints;

	// treat the first band differently, because the old gain values should
	// be overwritten. For the following bands, gain values will be added.
	filterList[0].calcCurve(curvePoints);

	// calculate the eq curve for a filter band
	for (int i = 1; i < (int)filterList.size(); i++) {
		filterList[i].calcCurve(tempCurve);

		// add the gain values to the summed curve
		for (int j = 0; j < (int)tempCurve.size(); j++) {
			curvePoints[j].setY(curvePoints[j].y() + tempCurve[j].y());
		}
	}
}

void JmbEq::reset()
{
	setOutputGain(0.0);

	if (hasActiveBand) {
		activeBand->setGain(0.0);
		if (activeBand->mode() == 6) 
			activeBand->setBandwidth(1.0);
		if ((activeBand->mode() == 7) || (activeBand->mode() == 8))
			activeBand->setBandwidth(0.7);
		if ((activeBand->mode() == 0) || (activeBand->mode() == 1))
			activeBand->setBandwidth(2.0);
	} else {
		for (int i = 0; i < (int)filterList.size(); i++) {
			filterList[i].setGain(0.0);
			if (filterList[i].mode() == 6) filterList[i].setBandwidth(1.0);
			if ((filterList[i].mode() == 7) || (filterList[i].mode() == 8))
				filterList[i].setBandwidth(0.7);
			if ((filterList[i].mode() == 0) || (filterList[i].mode() == 1))
				filterList[i].setBandwidth(2.0);
		}
	}

	calcEqCurve();
	update();
}

void JmbEq::toggleMode()
{
	if (!hasActiveBand) return;

	QList<int> modes;

	if (activeBand->id() == 0) {  // the lower end band
		modes << 6 << 7 << 1;
		int idx = modes.indexOf(activeBand->mode());
		if (idx == (modes.size() - 1)) {
			activeBand->setMode(modes[0]);
			activeBand->setBandwidth(1.0);
		} else {
			activeBand->setMode(modes[idx+1]);
			if (modes[idx+1] == 7)
				activeBand->setBandwidth(0.7);
			if (modes[idx+1] == 1)
				activeBand->setBandwidth(0.7);
		}
	}
	if (activeBand->id() == 3) {  // the upper end band
		modes << 6 << 8 << 0;
		int idx = modes.indexOf(activeBand->mode());
		if (idx == (modes.size() - 1)) {
			activeBand->setMode(modes[0]);
			activeBand->setBandwidth(1.0);
		} else {
			activeBand->setMode(modes[idx+1]);
			if (modes[idx+1] == 8)
				activeBand->setBandwidth(0.7);
			if (modes[idx+1] == 0)
				activeBand->setBandwidth(0.7);
		}
	}

	calcEqCurve();
	update();
}

/**                                                               **
 *    Coordinate transformation functions:                         *
 *    xToFreq converts widget x-coordinates to frequency values    *
 *    yToLevel converts widget y-coordinates to level values       *
 *    freqToX converts frequency values to widget x-coordinates    *
 *    levelToY converts level values to widget y-coordinates       *
 **                                                               **/

float JmbEq::xToFreq(int x)
{
	float ll = log10(LOWER_FREQ_LIMIT);
	float ul = log10(UPPER_FREQ_LIMIT);
	float range = ul - ll;
	float d = range * float(x) / float(width()-RMARGIN);
	float out = pow(10, d+ll);
	if (out > UPPER_FREQ_LIMIT) out = UPPER_FREQ_LIMIT;
	if (out < LOWER_FREQ_LIMIT) out = LOWER_FREQ_LIMIT;
	return out;
}

float JmbEq::yToLevel(int y)
{
	int range = UPPER_LEVEL_LIMIT - LOWER_LEVEL_LIMIT;
	float d = float(y - VMARGIN) / float(height() - 2*VMARGIN);
	return (float)UPPER_LEVEL_LIMIT - d * (float)range;
}

float JmbEq::f_freqToX(float f)
{
	float fl = log10(f);
	float ll = log10(float(LOWER_FREQ_LIMIT));
	float ul = log10(float(UPPER_FREQ_LIMIT));
	float range = ul - ll;
	float d = fl - ll;
	return (d / range) * float(width() - RMARGIN);
}

int JmbEq::freqToX(float f)
{
	return int(0.5 + f_freqToX(f));
}

float JmbEq::f_levelToY(float l)
{
	int range = UPPER_LEVEL_LIMIT - LOWER_LEVEL_LIMIT;
	float d = float(UPPER_LEVEL_LIMIT) - l;
	return (d / (float)range) * float(height() - 2.0*VMARGIN) + (float)VMARGIN;
}

int JmbEq::levelToY(float l)
{
	return int(0.5 + f_levelToY(l));
}

/* EOF */
